import torch
import torch.nn as nn

from enchanter.utils import backend as bf


__all__ = [
    "CausalConv1d"
]


class CausalConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, dilation=1, activation=torch.relu, **kwargs):
        """
        Causal Conv 1d

        Warnings:
            `torch.jit` を用いたJust-In-Time Compile には現在対応していません。

        Args:
            in_channels: 入力チャンネル数
            out_channels: 出力チャンネル数
            kernel_size: カーネルサイズ
            dilation: dilation
            activation: 活性化関数
            **kwargs:

        """
        super().__init__()
        self.dilation = dilation
        self.kernel_size = kernel_size
        self.padding = dilation * (kernel_size - 1)
        self.activation = activation

        self.conv1d = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            dilation=dilation,
            padding=self.padding,
            **kwargs
        )

    def forward(self, x):
        """
        順伝搬処理


        Args:
            x: [N, C, L] の３次元配列

        Returns:
            CausalConv1dを適用した結果 [N, C, L]

        """
        out = self.conv1d(x)
        if self.activation is not None:
            out = self.activation(out)

        if self.kernel_size > 0:
            out = bf.slice_axis(out, axis=2, begin=0, end=-self.padding)

        return out



import numpy as np
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split

import enchanter.wrappers as wrappers
import enchanter.addons.layers as layers
from enchanter.addons import Mish
from enchanter.callbacks import TensorBoardLogger
from enchanter.engine.modules import get_dataset


x, y = load_boston(return_X_y=True)
y = y.reshape(-1, 1)
x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, random_state=0)

train_ds = get_dataset(x_train.astype(np.float32), y_train.astype(np.float32))
val_ds = get_dataset(x_val.astype(np.float32), y_val.astype(np.float32))
test_ds = get_dataset(x_test.astype(np.float32), y_test.astype(np.float32))

train_loader = DataLoader(train_ds, batch_size=32, shuffle=True)
val_loader = DataLoader(val_ds, batch_size=32, shuffle=False)
test_loader = DataLoader(test_ds, batch_size=32, shuffle=False)

model = layers.MLP([13, 512, 128, 1], Mish())
optimizer = optim.Adam(model.parameters())


def test_regression_1():
    runner = wrappers.RegressionRunner(
        model,
        optimizer,
        nn.MSELoss(),
        TensorBoardLogger("./logs"),
    )
    runner.add_loader("train", train_loader).add_loader("val", val_loader).add_loader("test", test_loader)
    runner.train_config(epochs=1)

    try:
        runner.run(verbose=True)
        is_pass = True
    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass is True


def test_regression_2():
    runner = wrappers.RegressionRunner(
        model,
        optimizer,
        nn.MSELoss(),
        TensorBoardLogger("./logs")
    )
    runner.train_config(epochs=1)

    try:
        runner.run(verbose=False)
        is_pass = True

    except Exception:
        is_pass = False

    assert is_pass is False


def test_regression_3():
    runner = wrappers.RegressionRunner(
        model,
        optimizer,
        nn.MSELoss(),
        TensorBoardLogger("./logs")
    )
    try:
        runner.fit(x.astype(np.float32), y.astype(np.float32), batch_size=32, epochs=1)
        is_pass = True

    except Exception:
        is_pass = False

    assert is_pass is True



import numpy as np
from sklearn.metrics import roc_curve
import enchanter.metrics as metrics


def test_eer_1():
    # FRR = FN / (FN + TP)  ... FNR
    # FAR = FP / (TN + FP)  ... FPR = FP / (FP + TN)
    y_true = np.array([0, 0, 0, 0, 1, 1, 1, 1]).astype(np.float32)
    y_score = np.array([0.2, 0.3, 0.6, 0.8, 0.4, 0.5, 0.7, 0.9]).astype(np.float32)
    far, tpr, thresholds = roc_curve(y_true, y_score)

    frr = 1.0 - tpr
    idx = np.argmin(np.abs(far - frr))
    eer_from_frr_and_far = np.abs(far[idx] + frr[idx]) / 2
    eer = np.round(metrics.calculate_eer(y_true, y_score), 2)

    assert eer_from_frr_and_far == eer
